/* The code of the third glue example */
/* An example with a templates        */

#include <iostream>

template<typename T>
class GaussLike {
  T re;
  T im;
public:
  GaussLike () : re(0), im(0) {}
  GaussLike (T r, T i) : re(r), im(i) {}
  int real() { return re; }
  int imag() { return im; }
  void set_real(T r) { re = r; }
  void set_imag(T i) { im = i; }

  friend T real(const GaussLike& z) { return z.re; }
  friend T imag(const GaussLike& z) { return z.im; }
  friend int hash(const GaussLike& z) { return (hash(z.im) + hash(z.re)); }
};

template<typename T> inline 
bool operator == (const GaussLike<T>& a, const GaussLike<T>& b) {
  return (real(a) == real(b) ? (imag(a) == imag(b) ? true : false) : false);
}

template<typename T> inline
bool  operator != (const GaussLike<T>& a, const GaussLike<T>& b) {
  return ( !(a == b) );
}

template<typename T> GaussLike<T> 
operator+ (const GaussLike<T>& x, const GaussLike<T>& y) {
  return GaussLike<T> (real(x) + real(y), imag(x) + imag(y));
}

template<typename T> GaussLike<T> 
operator- (const GaussLike<T>& x, const GaussLike<T>& y) {
  return GaussLike<T> (real(x) - real(y), imag(x) - imag(y));
}

template<typename T> GaussLike<T> 
operator* (const GaussLike<T>& x, const GaussLike<T>& y) {
  return GaussLike<T> (real(x) * real(y) - imag(x) * imag(y),
                       real(x) * imag(y) + imag(x) * real(y));
}

template<typename T> std::ostream& 
operator<< (std::ostream& os, const GaussLike<T> z) {
  return os << "[ " << real(z) << "," << imag(z) << " ]";
}
