
/******************************************************************************
* MODULE     : mmx_extend.cc
* DESCRIPTION: generate generic extension code for C++ template libraries
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <symbols.hh>
#include <file.hh>
#include <fstream>
using std::fstream;
using std::ios;
using namespace mmx;

GC_GLOBAL_STATIC (bool, verbose, true);
GC_GLOBAL_STATIC (bool, so_flag, true);
GC_GLOBAL_STATIC (bool, glue_flag, true);
GC_GLOBAL_STATIC (bool, deps_only, false);
GC_GLOBAL_STATIC (string, output_dir, ".");
GC_GLOBAL_STATIC (string, next_output, "");
GC_GLOBAL_STATIC (string, local_lib_path, "");
GC_GLOBAL_STATIC (string, user_lib_path, "");
GC_GLOBAL_STATIC (string, last_name, "");
GC_GLOBAL_STATIC (string, compiler_option, "");
GC_GLOBAL_STATIC (string, linker_option, "");
GC_GLOBAL_STATIC (generic, mmh_prog, generic ());
GC_GLOBAL_STATIC (string, the_prefix, "");

string
get_root (string name) {
  int i, n= N(name);
  for (i=n-1; i>=0; i--)
    if (name[i] == '/') break;
  name= name (i+1, n);
  if (starts (name, "mmx-")) name= name (4, N(name));
  n= N(name);
  for (i=0; i<n; i++)
    if (name[i] == '.') break;
  return name (0, i);
}

string
get_out_name (string name) {
  if ((next_output != "") && (!so_flag)) return next_output;
  if (ends (name, ".mmh")) return name (0, N(name)-4) * ".cpp";
  return get_root (name) * ".cpp";
}

string
get_lib_name (string name) {
  if (next_output != "") return next_output;
  if (ends (name, ".mmh")) name= name (0, N(name)-4);
  if (output_dir == ".") return "mmx-" * get_root (name) * ".so";
  return output_dir * "/mmx-" * get_root (name) * ".so";
}

void
produce_dependencies () {
  string out_name= get_out_name (last_name);
  print_dependencies (out_name, the_prefix, mmh_prog);
}

void
produce_extension () {
  string out_name= get_out_name (last_name);
  string lib_name= get_lib_name (last_name);
  string root    = get_root (out_name);
  if (so_flag) root= get_root (lib_name);
  if (the_prefix != "") root= the_prefix;

  if (true) {
    // ostream& fout= cout; (void) out_name;
    char* _out_name= as_charp (out_name);
    fstream fout (_out_name, ios::out);
    mmx_free_charp (_out_name);
    if (verbose) cout << "mmx-extend: producing " << out_name << lf;
    if (extension_code (fout, root, mmh_prog, glue_flag))
      exit (1);
  }
  if (so_flag) {
    string cmd = string ("g++")
      * " -I" * sys_include_path 
      * " -I" * sys_include_path * "/mmx"
      * compiler_option
      * " " * out_name
      * " -o " * lib_name
      * " -rdynamic -fPIC -shared"
      * " -L" * user_lib_path
      * " -L" * local_lib_path
      * linker_option;
    if (verbose) cout << "mmx-extend: " << cmd << lf;
    if (system (cmd)) exit (1);
    cmd = "rm -f " * out_name;
    if (verbose) cout << "mmx-extend: " << cmd << lf;
    if (system (cmd)) exit (1);
  }
  next_output= "";
  so_flag= true;
  glue_flag= true;
}

void
process_option (char** args, int n, int& i) {
  string option = args[i];
  if (starts (option, "--")) option= option (1, N(option));
  if ((option == "-h") || (option == "-help")) {
    cout << "Options for mmx-extend:\n";
    cout << "  -c       : produce cpp file only\n";
    cout << "  -cflags  : print additional compiler flags\n";
    cout << "  -d dir   : specify output directory\n";
    cout << "  -g       : produce generic code only and not the glue\n";
    cout << "  -h       : display this help message\n";
    cout << "  -l       : equivalent to -d " << local_lib_path << "\n";
    cout << "  -libs    : print additional linking flags\n";
    cout << "  -M       : generate and print dependencies\n";
    cout << "  -s       : silent execution\n";
    cout << "  -o file  : specify the output file\n";
    cout << "  -p id    : specify a project prefix for separate compilation\n";
    cout << "  -u       : equivalent to -d " << user_lib_path << "\n";
    cout << "  -v       : display the version\n";
    cout << "  -Wc,opts : pass a list of options to the compiler\n";
    cout << "  -Wl,opts : pass a list of options to the linker\n";
    cout << "Please report bugs to <bugs@mathemagix.org>\n";
    exit (0);
  }
  if ((option == "-s") || (option == "-silent"))
    verbose= false;
  if ((option == "-o") || (option == "-output"))
    if (i+1 < n) next_output= args[++i];
  if ((option == "-d") || (option == "-dir"))
    if (i+1 < n) output_dir= args[++i];
  if ((option == "-l") || (option == "-local-libdir"))
    output_dir= local_lib_path;
  if ((option == "-u") || (option == "-user-libdir"))
    output_dir= user_lib_path;
  if ((option == "-v") || (option == "-version")) {
    cout << VERSION << "\n"; exit (0); }
  if ((option == "-c") || (option == "-cpp-code-only"))
    so_flag= false;
  if ((option == "-g") || (option == "-generic-code-only")) {
    so_flag= false; glue_flag= false; }
  if ((option == "-p") || (option == "-prefix"))
    if (i+1 < n) the_prefix= args[++i];
  if (starts (option, "-Wc,") || starts (option, "-Wl,")) {
    string s= option (4, N(option));
    for (int i=0; i<N(s); i++)
      if (s[i] == ',') s[i]= ' ';
    if (option[2] == 'c') compiler_option= compiler_option * " " * s;
    linker_option= linker_option * " " * s;
  }
  if (option == "-cflags") {
    cout << " -I" << sys_include_path
	 << " -I" << sys_include_path << "/mmx" << lf;
    exit (0); }
  if (option == "-libs") {
    cout << " -rdynamic -fPIC -shared"
	 << " -L" << user_lib_path
	 << " -L" << local_lib_path << lf;
    exit (0); }
  if (option == "-M")
    deps_only= true;
  if (option == "-test") {
    cout << "MMXDATA   = " << MMXDATA << "\n";
    cout << "MMXINCLUDE= " << MMXINCLUDE << "\n";
    cout << "MMXLIB    = " << MMXLIB << "\n";
    cout << "MMXSRC    = " << MMXSRC << "\n";
    exit (0);
  }
}

int
main (int argc, char** argv) {
  init_system ();
  mmh_prog= Begin ();
  local_lib_path= sys_lib_path;
  user_lib_path= sys_home * "/.mathemagix/lib";

  if (argc < 2) { 
    cout << "Usage: " << argv[0] << " --help\n";
    cout << "       " << argv[0] << " [options] files\n";
  }
  else {
    int i;
    for (i=1; i<argc; i++) {
      string arg = argv[i];
      if (starts (arg, "-")) process_option (argv, argc, i);
      else {
	generic prog= list (symbol ("load"), symbol (argv[i]));
	mmh_prog= Begin (cdr (as_list (mmh_prog)) * list (prog));
	last_name= argv[i];
      }
    }
    if (deps_only) produce_dependencies ();
    else produce_extension ();
  }
  return 0;
}
