
/******************************************************************************
* MODULE     : table.hh
* DESCRIPTION: tables
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __TABLE_HH
#define __TABLE_HH
#include <gentable.hh>
namespace mmx {

class table_rep: public gentable_rep {
protected:
  int size;      // size of table (nr of entries)
  int n;         // nr of keys (a power of two)
  list* a;       // the array of entries
  generic init;  // initial value of uninitialized arguments

protected:
  size_t Size ();
  void Mark ();
  int Type ();
  int Hash ();
  bool Equal (generic g);
  generic Copy ();
  int Arity ();
  void Print (ostream& out);

public:
  inline table_rep (int n2, generic init2):
    size (0), n (n2), a (new list [n]), init (init2) {}
  inline ~table_rep () { delete[] a; }

  void     resize (int n);
  generic  get_default ();
  generic  get_binding (generic key);
  void     set_binding (generic binding);
  void     reset (generic g);
  list     entries ();
  bool     contains (generic g);
  generic  read (generic g);
  generic& write (generic g);

  friend class table;
  friend int N (table a);
  friend bool operator == (table a, table b);
  friend table copy (table t);
  friend void print_entries (ostream& out, table t);
  friend void gc_mark_roots ();
  friend void gc_mark_sources ();
  friend void gc_memory_dump ();
};

class table {
DEFAULT_METHODS(table);
public:
  inline table (int n=2, generic init=generic()):
    rep (new table_rep (n, init)) {}
  inline operator gentable () { return gentable (rep); }

  inline generic  operator [] (generic g) { return rep->read (g); }
  inline generic& operator () (generic g) { return rep->write (g); }
  friend bool operator == (table a, table b);
  inline friend bool operator != (table a, table b) { return !(a==b); }

  inline friend int N (table t) { return t.rep->size; }
  friend table copy (table a);
  friend list entries (table a);
  friend table table_from_entries (list l);
  friend ostream& operator << (ostream& out, table a);
};

generic substitute (generic g, table t);
generic substitute (generic g, generic what, generic by);

}
#endif // __TABLE_HH
